CON

  SINE          = $e000

OBJ

 u      : "umath"

PUB distance_m(lat0, long0, lat1, long1) | dx, dy

  'Arguments are in millionths of a degree.
  'Returns distance in whole meters between latlong0 and latlong1.

  dy := lat1 - lat0
  dx := (long1 - long0) * cos(lat0 / 100_000) / 10_000
  return u.multdiv(^^(dx * dx + dy * dy),  111_319, 1_000_000)

PUB distance_mm(lat0, long0, lat1, long1) | dx, dy

  'Arguments are in millionths of a degree.
  'Returns distance in whole millimeters between latlong0 and latlong1.

  dy := lat1 - lat0
  dx := (long1 - long0) * cos(lat0 / 100_000) / 10_000
  return u.multdiv(^^(dx * dx + dy * dy),  111_319, 1_000)

PUB distance_cm(lat0, long0, lat1, long1) | dx, dy

  'Arguments are in millionths of a degree.
  'Returns distance in whole centimeters between latlong0 and latlong1.

  dy := lat1 - lat0
  dx := (long1 - long0) * cos(lat0 / 100_000) / 10_000
  return u.multdiv(^^(dx * dx + dy * dy),  111_319, 10_000)

PUB bearing(lat0, long0, lat1, long1) | dx, dy

  'Arguments are in millionths of a degree.
  'Returns bearing from latlong0 to latlong1 in tenths of a degree.

  dy := lat1 - lat0
  dx := (long1 - long0) * cos(lat0 / 100_000) / 10_000
  result := 900 - atan2(dy, dx)
  repeat while result < 0
    result += 3600
  result //= 3600

PUB degminsec2deg(degrees, minutes, second_hundredths)

  'Arguments are in whole degrees, whole minutes, and whole seconds.
  'Return value is in degrees * 1_000_000.

  return degrees * 1_000_000 + minutes * 100_000 / 6 + second_hundredths * 100 / 36

PUB degmin2deg (degrees, minute_hundred_thousandths)

  'Arguments are whole degrees, minutes * 100_000.
  'Return value is in degrees * 1_000_000.

  return degrees * 1_000_000 + minute_hundred_thousandths / 6

PUB cos(x)

  'Argument x is in tenths of a degree.

  return sin(x + 900)

PUB sin(x) : value

  'Argument x is in tenths of a degree.

  x := x * $2000 / 3600

  if (x & $fff == $800)
    value := $1_0000
  elseif (x & $800)
    value := word[SINE][-x & $7ff]
  else
    value := word[SINE][x & $7ff]
  if (x & $1000)
    value := -value
  value := (value * 10_000) ~> 16

PUB atan2(y, x) : a | negate, i, da, dx, dy

  'CORDIC atan2 algorithm
  '
  'Arguments are in any identcally-scaled units.
  'Hypotenuse of x,y must be within 1_300_000_000 to avoid overflow
  'Return value is in tenths of a degree between -1800 and 1799.

  i := 29 - >|(||x #> ||y)    'Make the arguments as large as possible.
  if (i > 0)
    x <<= i
    y <<= i
     
  a~    
  negate := x < 0             'check quadrant 2 | 3                        
   
  if negate                   'if quadrant 2 | 3, (may be outside 106 convergence range)                                                                 
    a += $80000000            '..add 180 degrees to angle
    -x                        '..negate x
    -y                        '..negate y

  repeat i from 0 to 26       'do CORDIC iterations (27 is optimal for 32-bit values)
    da := corlut[i]
    dx := y ~> i
    dy := x ~> i

    negate := y < 0           'if atan2 mode, drive y towards 0

    if negate
      -da
      -dx
      -dy

    a += da
    x += dx
    y -= dy

  a := (((a ~> 16) * 3600) ~> 16) 'Convert to tenths of a degree.

PUB error_angle(angle)

  'Argument is in thenths of a degree.
  'Result is the same angle translated to lie between -1800 and 1800.

  result := angle
  repeat while (result > 1800)
    result -= 3600
  repeat while (result < -1800)
    result += 3600

PUB compass_angle(angle)

  'Computes compass angle (North = 0, increasing clockwise.) from
  'trig angle (East = 0, increasing counter-clockwise).
  'Units are tenths of a degree.

  result := angle
  repeat while result > 3600
    result -= 3600
  repeat while result < 0
    result += 3600

DAT

corlut        long $20000000    'CORDIC angle lookup table
              long $12E4051E
              long $09FB385B
              long $051111D4
              long $028B0D43
              long $0145D7E1
              long $00A2F61E
              long $00517C55
              long $0028BE53
              long $00145F2F
              long $000A2F98
              long $000517CC
              long $00028BE6
              long $000145F3
              long $0000A2FA
              long $0000517D
              long $000028BE
              long $0000145F
              long $00000A30
              long $00000518
              long $0000028C
              long $00000146
              long $000000A3
              long $00000051
              long $00000029
              long $00000014
              long $0000000A